@@ -123,7 +123,13 @@ spi_dec_plat_mouse_check (SpiDEController *controller,
   if (klass->plat.mouse_check)
     return klass->plat.mouse_check (controller, x, y, moved);
   else
-    return 0;
+    {
+      if (moved)
+        {
+          *moved = FALSE;
+        }
+      return 0;
+    }
 }
 
 static gboolean
@@ -1675,6 +1681,7 @@ impl_get_device_event_listeners (DBusConnection *bus,
 static unsigned
 get_modifier_state (SpiDEController *controller)
 {
+	spi_dec_poll_mouse_moved (controller);
 	return mouse_mask_state;
 }
 
@@ -1696,12 +1703,17 @@ spi_dec_synth_keysym (SpiDEController *controller, long keysym)
 	if (synth_mods != modifiers) {
 		lock_mods = synth_mods & ~modifiers;
 		spi_dec_plat_lock_modifiers (controller, lock_mods);
+		if (modifiers & LockMask)
+			spi_dec_plat_unlock_modifiers (controller, LockMask);
 	}
 	spi_dec_plat_synth_keycode_press (controller, key_synth_code);
 	spi_dec_plat_synth_keycode_release (controller, key_synth_code);
 
-	if (synth_mods != modifiers) 
+	if (synth_mods != modifiers) {
 		spi_dec_plat_unlock_modifiers (controller, lock_mods);
+		if (modifiers & LockMask)
+			spi_dec_plat_lock_modifiers (controller, LockMask);
+	}
 	return TRUE;
 }
 
